# accapi/forms.py
import requests
from datetime import date, timedelta
from .base import AccBase

class AccFormsApi:
    '''
    This class provides methods to interact with the Forms endpoint of the Autodesk Construction Cloud API.
    The GET methods require data:read and the POST, PATCH, and PUT methods require data:write scope.

    Token must be Bearer <token>, where <token> is is obtained via a three-legged OAuth flow
    for the GET, POST, PATCH, and PUT methods.
    '''       
    def __init__(self, base: AccBase):
        self.base = base        
        self.base_url = "https://developer.api.autodesk.com/construction/forms/v1"

    def get_forms(self, project_id, offset=0, limit=50, ids=None, formDateMin=None,
                  formDateMax=None, updatedAfter=None, updatedBefore=None, templateId=None,
                  statuses=None, sortBy=None, sortOrder=None, locationIds=None, follow_pagination=False)->list[dict]:
        """
        Returns a paginated list of forms in a project. Forms are sorted by updatedAt, most recent first.
        Returns up to 50 forms per page unless follow_pagination is True.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/forms-forms-GET/

        Args:
            project_id (str): The project ID.
            offset (int): The number of records to skip (defaults to 0).
            limit (int): The number of records per page (defaults to 50, maximum is 50).
            ids (str, optional): A comma-separated list of form IDs to filter by.
            formDateMin (str, optional): ISO 8601 datetime to filter forms created after this time.
            formDateMax (str, optional): ISO 8601 datetime to filter forms created before this time.
            updatedAfter (str, optional): ISO 8601 datetime to filter forms updated after this time.
            updatedBefore (str, optional): ISO 8601 datetime to filter forms updated before this time.
            templateId (str, optional): The form template ID to filter by.
            statuses (str, optional): A comma-separated list of statuses to filter by.
            sortBy (str, optional): The field to sort the forms by.
            sortOrder (str, optional): The sort order ('asc' or 'desc').
            locationIds (str, optional): A comma-separated list of location IDs to filter by.
            follow_pagination (bool, optional): Whether to follow pagination links.

        Returns:
            list: A list of forms in the project.
        """        
        # Remove the "b." prefix if present.
        if project_id.startswith("b."):
            project_id = project_id[2:]
        
        # Build the initial URL
        url = f"{self.base_url}/projects/{project_id}/forms"
        
        # Build the query parameters dictionary
        params = {
            "offset": offset,
            "limit": limit,
        }
        if limit > 50:
            raise ValueError("The maximum limit is 50.")
        
        if ids is not None:
            params["ids"] = ids
        if formDateMin is not None:
            params["formDateMin"] = formDateMin
        if formDateMax is not None:
            params["formDateMax"] = formDateMax
        if updatedAfter is not None:
            params["updatedAfter"] = updatedAfter
        if updatedBefore is not None:
            params["updatedBefore"] = updatedBefore
        if templateId is not None:
            params["templateId"] = templateId
        if statuses is not None:
            params["statuses"] = statuses
        if sortBy is not None:
            params["sortBy"] = sortBy
        if sortOrder is not None:
            params["sortOrder"] = sortOrder
        if locationIds is not None:
            # Autodesk expects multiple location parameters as individual query parameters.
            params["locationId"] = locationIds

        headers = {"Authorization": f"Bearer {self.base.get_3leggedToken()}"}

        all_forms = []        
        current_params = params

        while url:
            response = requests.get(url, headers=headers, params=current_params)
            response.raise_for_status()
            json_data = response.json()

            # Extend the list with the forms data from this page.
            all_forms.extend(json_data.get("data", []))

            # Get pagination info
            pagination = json_data.get("pagination", {})
            url = pagination.get("nextUrl") if follow_pagination else None

            # For subsequent requests, use the next_url which already includes query parameters.
            current_params = {}

        return all_forms

    def get_templates(self, project_id, offset=0, limit=50, updatedAfter=None,
                      updatedBefore=None, sortOrder=None, follow_pagination=False)->list[dict]:
        """
        Retrieve all form templates for a given project by following pagination.
        
        https://aps.autodesk.com/en/docs/acc/v1/reference/http/forms-form-templates-GET/
        
        Args:
            project_id (str): The project ID (will remove a leading "b." if present).
            offset (int): Number of records to skip (defaults to 0).
            limit (int): Number of records per page (defaults to 50, maximum is 50).
            updatedAfter (str, optional): ISO 8601 datetime to filter templates updated after this time.
            updatedBefore (str, optional): ISO 8601 datetime to filter templates updated before this time.
            sortOrder (str, optional): Sort order for the templates ('asc' or 'desc').

        Returns:
            list: A list of all form templates for the specified project.
        """
        # Remove the "b." prefix from the project_id if it exists.
        if project_id.startswith("b."):
            project_id = project_id[2:]

        # Build the initial URL for the form templates endpoint.
        url = f"{self.base_url}/projects/{project_id}/form-templates"

        # Build the query parameters.
        params = {
            "offset": offset,
            "limit": limit,
        }
        if limit > 50:
            raise ValueError("The maximum limit is 50.")
        if updatedAfter is not None:
            params["updatedAfter"] = updatedAfter
        if updatedBefore is not None:
            params["updatedBefore"] = updatedBefore
        if sortOrder is not None:
            params["sortOrder"] = sortOrder

        headers = {"Authorization": f"Bearer {self.base.get_3leggedToken()}"}

        all_templates = []        
        current_params = params

        while url:
            response = requests.get(url, headers=headers, params=current_params)
            response.raise_for_status()
            json_data = response.json()

            # Add the templates from this page to the list.
            all_templates.extend(json_data.get("data", []))

            # Follow the pagination link if available.
            pagination = json_data.get("pagination", {})
            url = pagination.get("nextUrl") if follow_pagination else None
            if not url:
                break

            current_params = {}

        return all_templates

    def get_forms_for_past30(self, project_id, **kwargs)->list[dict]:
        """
        Retrieve forms for a given project that were created within teh past 30 days.

        Args:
            project_id (str): The project ID.
            **kwargs: Additional keyword arguments to pass to the get_forms() method.

        Returns:
            list: A list of forms that were created within the past 30 days.            
        """
        today = date.today()
        date_minus_30 = today - timedelta(days=30)

        # Format the date as 'yyyy-mm-dd'
        formDateMin = date_minus_30.strftime('%Y-%m-%d')

        return self.get_forms(project_id, formDateMin=formDateMin, **kwargs)

    def post_form(self, project_id:str, template_id:str, data:dict)->dict:
        """
        Adds a new form to a project.
        
        https://aps.autodesk.com/en/docs/acc/v1/reference/http/forms-forms-POST/

        Args:
            project_id (str): The project ID from GET forms.
            template_id (str): The form template ID from GET forms.
            data (dict): The form data to create the form.

        """
        url = f"{self.base_url}/projects/{project_id}/form-templates/{template_id}/forms"

        headers = {"Authorization": f"Bearer {self.base.get_3leggedToken()}", "Content-Type": "application/json"}
        response = requests.post(url, headers=headers, json=data)
        response.raise_for_status()
        return response.json()

    def patch_form(self, project_id:str, template_id:str, form_id:str, data:dict)->dict:
        """
        Updates a form’s form details. Note that we do not currently support updating PDF forms.

        To edit a form, it must be in draft or inReview status and the user must have permissions to edit the form.
        
        https://aps.autodesk.com/en/docs/acc/v1/reference/http/forms-forms-formId-PATCH/

        Args:
            project_id (str): The project ID from GET forms.            
            template_id (str): The form template ID from GET forms.
            form_id (str): The form ID from GET forms.
            data (dict): The form data split into customValues(non-tabular fields) and tabularValues(tabular fields).

        """
        url = f"{self.base_url}/projects/{project_id}/form-templates/{template_id}/forms/{form_id}"

        headers = {"Authorization":f"Bearer {self.base.get_3leggedToken()}", 
                   "Content-Type": "application/json"}
        response = requests.patch(url, headers=headers, json=data)
        response.raise_for_status()
        return response.json()

    def put_form(self, project_id:str, form_id:str, data:dict)->dict:
        """
        Updates a form’s main form fields, both tabular and non-tabular. Note that we do not currently support updating PDF forms.
        To edit form values, the form needs to be in draft status and the user must have permissions to edit the form. 
        
        https://aps.autodesk.com/en/docs/acc/v1/reference/http/forms-forms-formId-PUT/

        Args:
            project_id (str): The project ID from GET forms.            
            data (dict): The form data split into customValues(non-tabular fields) and tabularValues(tabular fields).

        Returns:
            dict: The form object from the API response.
        """
        url = f"{self.base_url}/projects/{project_id}/forms/{form_id}/values:batch-update"

        headers = {"Authorization":f"Bearer {self.base.get_3leggedToken()}", 
                   "Content-Type": "application/json"}
        response = requests.put(url, headers=headers, json=data)
        response.raise_for_status()
        return response.json()


from calendar import c
import requests
from .base import AccBase


class AccBusinessUnitsApi:
    def __init__(self, base: AccBase):
        self.base_url = "https://developer.api.autodesk.com/hq/v1/accounts/:account_id/business_units_structure"
        self.base = base

    def get_business_units(self)->list[dict]:
        """
        Query all the business units in a specific BIM 360 account.
        Authorization must be Bearer <token>, where <token> is obtained via a two-legged OAuth flow.
        The GET method requires the account:read scope, and the Authentication Context	
        is app only.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/business_units_structure-GET/

        Returns:
            list[dict]: An array of business units associated with the account_id                
        """
        token = f"Bearer {self.base.get_2leggedToken()}"
        headers = {"Authorization": token}
        url = self.base_url.replace(":account_id", self.base.account_id)

        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            content = response.json()
            return content.get("business_units", [])
        else:
            error = response.json().get('errors')
            if error:
                print(error)
            detail = response.json().get('detail')
            if detail:
                print(detail)
            response.raise_for_status()

    def update_business_units(self, business_units:list[dict])->list[dict]:
        """
        Creates or redefines the business units of a specific BIM 360 account.
        Authorization must be Bearer <token>, where <token> is obtained via a two-legged OAuth flow.
        The PUT method requires the account:write scope, and the Authentication Context is app only.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/business_units_structure-PUT/
            
        Args:
            business_units (dict): The business units to create or redefine. Each business unit
            must have a unique name, and optional id, parent_id, and description fields.
            If Id is specified and already existing, the existing business unit will be replaced
            with the provided attributes. If ID specified and not already existing, a new business 
            unit will be created with the id. If Id unspecified, a new business unit will be created
            with a server-generated id.

        Returns:
            list[dict]: An array of created or modified business units.
        """
        token = f"Bearer {self.base.get_2leggedToken()}"
        headers = {"Authorization": token,
                   "Content-Type": "application/json"}
        url = self.base_url.replace(":account_id", self.base.account_id)

        response = requests.put(url, headers=headers, json=business_units)
        if response.status_code == 200:
            content = response.json()
            return content
        else:
            error = response.json().get('errors')
            if error:
                print(error)
            detail = response.json().get('detail')
            if detail:
                print(detail)
            response.raise_for_status()


from re import A
from .base import AccBase
from .account_users import AccAccountUsersApi
from .project_users import AccProjectUsersApi
from .projects import AccProjectsApi
from .sheets import AccSheetsApi
from .data_management import AccDataManagementApi
from .forms import AccFormsApi
from .authentication import Authentication
from .companies import AccCompaniesApi
from .business_units import AccBusinessUnitsApi

# Aggregator class supporting both 2-legged and 3-legged auth.
class Acc:
    def __init__(   self, 
                    auth_client: Authentication,
                    account_id=None):
        """
        Initializes the Acc aggregator.

        Parameters:
            auth_client (Authentication): The authentication client to use.
            account_id (str): The account ID to use. If not provided, the account ID is determined from the auth_client.

        """

        # Initialize the shared base instance.
        self.base = AccBase(auth_client=auth_client, account_id=account_id)
        self.auth_client = auth_client
        
        # Instantiate APIs that work with both token types.        
        self.project_users = AccProjectUsersApi(self.base)
        self.projects = AccProjectsApi(self.base)
        self.sheets = AccSheetsApi(self.base)
        self.data_management = AccDataManagementApi(self.base)      

        # Apis that require 2-legged authorization
        self.account_users = AccAccountUsersApi(self.base)
        self.business_units = AccBusinessUnitsApi(self.base)

        # Requires account:read scope and a 2-legged or 3-legged auth token for GET methods
        # Requires account:write scope and a 2-legged auth token for POST and PATCH methods
        self.companies = AccCompaniesApi(self.base)

        # Apis that require 3-legged authorization
        self.forms = AccFormsApi(self.base) 
        self.user_profiles = AccAccountUsersApi(self.base)
    
    def get_forms(self, project_id, **forms_kwargs)->list[dict]:
        """
        get_forms extension to get forms along with user names and email addresses
        of the users who created the forms.

        Args:
            project_id (str): The project ID.
            **forms_kwargs: Additional keyword arguments to pass to the get_forms() method.

        Returns:
            list: A list of forms along with user names and email addresses of the users who created the forms.
        """
        # Retrieve the form templates for the project.
        forms = self.forms.get_forms(project_id, **forms_kwargs)

        # Extract the createdBy field the forms and create a list of only unique names
        user_names = []
        for form in forms:
            user_name = form.get("createdBy")
            if user_name not in user_names:
                user_names.append(user_name)

        # get user names from the user ids using the project_users api
        project_users = self.project_users.get_users(project_id, query_params={"filter['autodeskId']": user_names,"fields":['email','name','autodeskId']})

        # update forms with user names and email addresses matched by form.createdBy = project_user.autodeskId
        for form in forms:
            userId = form.get("createdBy")
            for project_user in project_users:
                if userId == project_user.get("autodeskId"):
                    form["createdByName"] = project_user.get("name")
                    form["createdByEmail"] = project_user.get("email")
                    break

        return forms

    def get_forms_for_past30(self, project_id, **forms_kwargs)->list[dict]:
        """
        get_forms_for_past30 extension to get forms along with user names and email addresses
        of the users who created the forms that were created within the past 30 days.

        Args:
            project_id (str): The project ID.
            **forms_kwargs: Additional keyword arguments to pass to the get_forms() method.

        Returns:
            list: A forms entered over the past 30 days.
        
        """
        forms = self.forms.get_forms_for_past30(project_id, **forms_kwargs)
        
        # Extract the createdBy field the forms and create a list of only unique names
        user_names = []
        for form in forms:
            user_name = form.get("createdBy")
            if user_name not in user_names:
                user_names.append(user_name)

        # get user names from the user ids using the project_users api
        project_users = self.project_users.get_users(
            project_id, 
            query_params={"filter['autodeskId']": user_names,"fields":['email','name','autodeskId']})

        # update forms with user names and email addresses matched by form.createdBy = project_user.autodeskId
        for form in forms:
            userId = form.get("createdBy")
            for project_user in project_users:
                if userId == project_user.get("autodeskId"):
                    form["createdByName"] = project_user.get("name")
                    form["createdByEmail"] = project_user.get("email")
                    break

        return forms

    def get_forms_all_active_projects(self, **forms_kwargs):
        """
        get_forms_all_projects extension to get forms along with user names and email addresses
        of the users who created the forms for all projects.

        Args:
            **forms_kwargs: Additional keyword arguments to pass to the get_forms() method.
        
        """
        projects = self.projects.get_all_active_projects()
        all_forms = []
        for project in projects:
            forms = self.get_forms(project.get("id"), **forms_kwargs)
            all_forms.extend(forms)
        return all_forms

    def get_forms_all_active_projects_last30(self, **forms_kwargs):
        """
        get_forms_all_projects_last30 extension to get forms along with user names and email addresses
        of the users who created the forms for all projects that were created within the past 30 days.

        Args:
            **forms_kwargs: Additional keyword arguments to pass to the get_forms() method.
        
        """
        projects = self.projects.get_all_active_projects()
        all_forms = []
        for project in projects:
            forms = self.get_forms_for_past30(project.get("id"), **forms_kwargs)
            all_forms.extend(forms)
        return all_forms
        
    
        



import requests
from .base import AccBase

class AccProjectsApi:
    '''
    This class provides methods to interact with the project endpoint of the Autodesk Construction Cloud API.
    The get methods require account:read and the post, and patch methods require account:write scopes.

    Token must be Bearer <token>, where <token> is is obtained via either a two-legged or three-legged OAuth flow
    for the GET and POST methods.  The Patch method requires a 2-legged token.
    '''    
    def __init__(self, base: AccBase):
        self.base_url = "https://developer.api.autodesk.com/construction/admin/v1"
        self.base = base
        self.user_id = self.base.user_info.get('uid')

    def get_project(self, project_id):
        """
        Get a project by project ID.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-projectsprojectId-GET/

        Args:
            project_id (str): The project ID to lookup the project by.

        Returns:
            dict: project object
        """
        token = f"Bearer {self.base.get_private_token()}"

        headers = {"Authorization": token}
        url = f"https://developer.api.autodesk.com/construction/admin/v1/projects/{project_id}"

        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            content = response.json()
            return content
        else:
            error = response.json().get('errors')
            if error:
                print(error)
            detail = response.json().get('detail')
            if detail:
                print(detail)
            response.raise_for_status()

    def get_projects(self, filter_params:dict=None, follow_pagination:bool=False)->list[dict]:
        """
        Get a list of projects in an account. You can also use this endpoint to filter out the list of projects by setting the filter parameters.   

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-accounts-accountidprojects-GET/

        Args:
            filter_params (dict, optional): Filter parameters to apply to the query.
            follow_pagination (bool, optional): If True, follow pagination links to get all projects.

        Returns:
            list: The list of projects (results array) from the API response.
        """        
        token = f"Bearer {self.base.get_private_token()}"

        # if filter_params contains fields, and fields contains ID, then remove ID from fields
        if filter_params is not None and filter_params.get('fields') != None:
            fields = filter_params.get('fields').split(',')
            fields = [field.strip() for field in fields]
            if 'id' in fields:
                fields.remove('id')
                filter_params['fields'] = ','.join(fields)

        url = f"{self.base_url}/accounts/{self.base.account_id}/projects"

        headers = {"Authorization": token, "User-Id": self.user_id}

        projects = []
        while url:
            response = requests.get(url, headers=headers, params=filter_params)
            if response.status_code != 200:
                errors = response.json().get('errors') if response.json().get('errors') else response.json().get('detail')
                if errors:
                    print(errors)                
                response.raise_for_status()

            data = response.json()
            projects.extend(data.get("results",{}))
            
            url = data["pagination"].get("nextUrl") if follow_pagination else None

            # clear filter_params for subsequent requests using the pagination URL
            filter_params = None

        return projects

    def get_active_projects(self, filter_params:dict=None, follow_pagination:bool=False):
        """Get all active projects in the account.

        Args:
            filter_params (str, optional): Filter parameters to apply to the query. 
            follow_pagination (bool, optional): If True, follow pagination links to get all projects.

        Returns:
            list: a list of active projects filtered by the filter parameters.
        """
        if filter_params is None:
            filter_params = {}
        filter_params['filter[status]'] = 'active'
        return self.get_projects(filter_params, follow_pagination)

    def get_all_active_projects(self, filter_params:dict=None):
        """Get all active projects in the account.

        Args:
            filter_params (str, optional): Filter parameters to apply to the query.            

        Returns:
            list: a list of active projects filtered by the filter parameters.
        """        
        if filter_params is None:
            filter_params = {}
        filter_params['filter[status]'] = 'active'
        return self.get_projects(filter_params, follow_pagination=True)
    
    def post_project(self, project: dict):
        """
        Creates a new project in the specified account. You can create the project directly,
        or clone it from a project template.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-accounts-accountidprojects-POST/
        
        Args:
            project (dict): A project object to create. Required fields are name and type. 
            Name must be unique. 

        Returns:
            dict: The created project object.
        """
        token = f"Bearer {self.base.get_private_token()}"

        if project.get("name") == None:
            raise Exception("Missing required parameter 'name'")
        if project.get("type") == None:            
            raise Exception("Missing required parameter 'type'")

        headers = {
            "Content-Type": "application/json",
            "Authorization": token,
            "User-Id": self.user_id,
        }
        response = requests.post(
            f"{self.base_url}/accounts/{self.base.account_id}/projects",
            headers=headers,
            json=project,
        )
        if response.status_code == 202:
            print(f"Project {project['jobNumber']} created")
            content = response.json()
            return content
        elif response.status_code == 409:
            errors = response.json().get('errors') if response.json().get('errors') else response.json().get('detail')
            if errors:                
                [print(error) for error in errors]
            job = self.get_projects(filter_params={'filter[jobNumber]': project['jobNumber']})
            return job[0]
        else:
            errors = response.json().get('errors') if response.json().get('errors') else response.json().get('detail')
            if errors:                
                [print(error) for error in errors]
            response.raise_for_status()




# accapi/project_users.py
import requests
from .base import AccBase
import time


class AccProjectUsersApi:
    '''
    This class provides methods to interact with the project users endpoint of the Autodesk Construction Cloud API.
    The class provides methods to get, add, update, and delete users in a project as well as bulk implementaitons of these methods.
    
    Token must be Bearer <token>, where <token> is is obtained via either or with two-legged or three-legged oauth flow.
    The GET methods require account:read and the POST, PATCH, and DELETE methods require account:write scopes.
    The POST, PATCH, and DELETE methods require the User-Id header to be set to the user's ID when two-legged authentication
    is used.    
    '''

    productadmin = [
                    {
                        "key": "projectAdministration",
                        "access": "administrator"
                    },
                    {
                        "key": "designCollaboration",
                        "access": "administrator"
                    },
                    {
                        "key": "build",
                        "access": "administrator"
                    },
                    {
                        "key": "cost",
                        "access": "administrator"
                    },
                    {
                        "key": "modelCoordination",
                        "access": "administrator"
                    },
                    {
                        "key": "docs",
                        "access": "administrator"
                    },
                    {
                        "key": "insight",
                        "access": "administrator"
                    },
                    {
                        "key": "takeoff",
                        "access": "administrator"
                    }
                    ]

    productmember = [
                    {
                        "key": "projectAdministration",
                        "access": "none"
                    },
                    {
                        "key": "designCollaboration",
                        "access": "member"
                    },
                    {
                        "key": "build",
                        "access": "member"
                    },
                    {
                        "key": "cost",
                        "access": "member"
                    },
                    {
                        "key": "modelCoordination",
                        "access": "member"
                    },
                    {
                        "key": "docs",
                        "access": "member"
                    },
                    {
                        "key": "insight",
                        "access": "member"
                    },
                    {
                        "key": "takeoff",
                        "access": "member"
                    }
                    ]  
    
    def __init__(self, base: AccBase):
        self.base_url = "https://developer.api.autodesk.com/construction/admin"
        self.base = base
        self.user_id = self.base.user_info.get('uid')

    def get_users(self, project_id:str, query_params=None, follow_pagination=False):
        """
        Retrieves information about a filtered subset of users in the specified project.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-projectsprojectId-users-GET/

        Args:
            project_id (str): Project UUID
            query_params (dict, optional): One of the filters described in the official documentation.
            follow_pagination (bool, optional): Returns all results if True. If False returns up to 200 results.

        Returns:
            list[dict]: user information, filtered by query_params
        """
        token = f"Bearer {self.base.get_private_token()}"

        base_url = f"{self.base_url}/v1/projects/{project_id}/users"
        headers = {"Authorization": token, "User-Id": self.user_id}
        all_users = []
        next_url = base_url

        if query_params is None:
            query_params = {}
            
        # add limit, offset to query_params if they dont exist
        if query_params.get('limit') is None:
            query_params['limit'] = 200
        if query_params.get('offset') is None:
            query_params['offset'] = 0
        
        
        while next_url:
            response = requests.get(next_url, headers=headers, params=query_params)
            if response.status_code != 200:
                response.raise_for_status()

            data = response.json()            
            all_users.extend(data["results"])

            next_url = data["pagination"].get("nextUrl") if follow_pagination else None

            # Clear query parameters for subsequent requests using the pagination URL
            query_params = None

        # foreach user in all_users, add user['uid'] = user['autodeskId']
        for user in all_users:
            user["sub"] = user["uid"] = user["autodeskId"]
        
        return all_users

    def get_user(self, project_id:str, user_id:str, fields:list[str]=[])->dict:
        """
        Retrieves detailed information about the specified user in a project.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-projectsprojectId-users-userId-GET/    

        Args:
            project_id (str): The project UUID
            user_id (str): The ID of the user. You can use either the ACC ID (id) or the Autodesk ID (autodeskId).
            fields (list[str], optional): List of user fiuelds to return.

        Returns:
            dict: Detailed project user dictionary.
        """
        token = f"Bearer {self.base.get_private_token()}"

        url = f"{self.base_url}/v1/projects/{project_id}/users/{user_id}"
        headers = {"Authorization": token}
        query_params = {"fields": fields}
        response = requests.get(url, headers=headers, params=query_params)

        if response.status_code == 200:
            user = response.json()
            user["sub"] = user["uid"] = user["autodeskId"]
            return user
        else:
            response.raise_for_status()

    def add_user(self, project_id:str, user: dict)->dict:
        """
        Adds a user to a project.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-projects-project-Id-users-POST/      

        Args:
            project_id (str): _description_the UUID of the project.
            user (dict): A user dictionary including a minimum of email and products keys.

        Returns:
            dict: user dictionary    
        """
        url = f"{self.base_url}/v1/projects/{project_id}/users"

        headers = {
            "Authorization": f"Bearer {self.base.get_private_token()}",
            "Content-Type": "application/json",
            "User-Id": self.user_id,
        }

        if user.get("email") is None:
            raise ValueError("Email is required")
        if user.get("products") is None:
            raise ValueError("Products is required")

        modified_user = {"email": user.get("email"), "products": user.get("products")}
        if user.get("roleIds"):
            modified_user["roleIds"] = user.get("roleIds")

        response = requests.post(url, headers=headers, json=modified_user, timeout=100)
        response.raise_for_status()

        return response.json()
        
    def import_users(self, project_id:str, users: list[dict]):
        """
        Imports a list of users into a project.
        
        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-v2-projects-project-Id-users-import-POST/
        
        Args:
            project_id (str): A project UUID
            users (list[dict]): A list of user dictionaries, each containing email and products keys.

        Returns:
            None
        """
        token = f"Bearer {self.base.get_private_token()}"

        url = f"{self.base_url}/v2/projects/{project_id}/users:import"
        
        headers = {
            "Authorization": token,
            "Content-Type": "application/json",
            "User-Id": self.user_id
        }

        if len(users) == 0:
            print("No users to import")
            return

        # Modify users list to retain only specified fields and add roleIds
        modified_users = []
        for user in users:
            modified_user = {
                "email": user.get("email"),
                "products": user.get("products"),
                #'roleIds': [user.get('default_role_id')]
            }
            modified_users.append(modified_user)

        # Split the users list into chunks of 200
        for i in range(0, len(modified_users), 200):
            chunk = modified_users[i : i + 200]
            data = {"users": chunk}

            response = requests.post(url, headers=headers, json=data)
            if response.status_code == 202:
                print(f"Chunk {i//200 + 1} imported successfully")
            else:
                response.raise_for_status()

    def patch(self, project_id:str, target_user_id:str, data:dict):
        """
        Updates a user in a project.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-projects-project-Id-users-userId-PATCH/
        
        Args:
            project_id (str): A project UUID
            target_user_id (str): ID of user to update
            data (dict): User data to update  

        Returns:
            dict: a user dictionary
        """
        token = f"Bearer {self.base.get_private_token()}"

        url = f"{self.base_url}/v1/projects/{project_id}/users/{target_user_id}"
        
        headers = {
            "Authorization": token,
            "Content-Type": "application/json",
            "User-Id": self.user_id,
        }
        response = requests.patch(url, headers=headers, json=data)
        if response.status_code == 200:
            return response.json()
        else:
            response.raise_for_status()

    def patch_project_users(self, projects: list[dict], users: list[dict], products: list[dict] = None):
        """
        Patch users in a project using dictionaries.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-projects-project-Id-users-userId-PATCH/
        
        Args:
            projects (list[dict]): A list of project dictionaries.
            users (list[dict]): A list of user dictionaries.
            products (list[dict], optional): A list of product dictionaries. Defaults to productmember.
        
        Returns:
            None
        """
        # Use the provided products or the default if None.
        if products is None:
            products = productmember  # assuming productmember is defined globally
        
        # Build a dictionary of users keyed by email.
        user_dict = {user.get('email'): user for user in users if user.get('email')}
        
        for project in projects:
            project_id = project.get('id')
            if not project_id:
                continue  # Skip projects without an id
            
            # Assume self.get_all returns a list of project user dictionaries.
            project_users = self.get_all(project_id)
            project_user_dict = {pu.get('email'): pu for pu in project_users if pu.get('email')}
            
            for email, user in user_dict.items():
                if email in project_user_dict:
                    pu = project_user_dict[email]
                    # Retrieve the access level from the first product in the user's dictionary
                    user_products = user.get('products', [])
                    user_access = user_products[0].get('access') if user_products and isinstance(user_products, list) else None

                    # Check if any of the project user's products need updating
                    if any(prod.get('key') == 'projectAdministration' and prod.get('access') != user_access
                        for prod in pu.get('products', [])):
                        print(f"patching user {email} to project {project.get('jobNumber', 'unknown')}")
                        try:
                            self.patch(project_id, pu.get('id'), {'products': products})
                        except Exception:
                            time.sleep(60)
                            self.patch(project_id, pu.get('id'), {'products': products})
                    else:
                        print(f"user {email} already patched to project {project.get('jobNumber', 'unknown')}")


    def delete(self, project_id:str, target_user_id:str):
        """
        Removes the specified user from a project.

        https://aps.autodesk.com/en/docs/acc/v1/reference/http/admin-projects-project-Id-users-userId-DELETE/
        
        Args:
            project_id (str): A project UUID
            target_user_id (str): The ID of the user to remove from the project.

        Returns:
            None
        """
        token = f"Bearer {self.base.get_private_token()}"

        url = f"{self.base_url}/v1/projects/{project_id}/users/{target_user_id}"

        headers = {"Authorization": token, 
                   "User-Id": self.user_id}

        response = requests.delete(url, headers=headers)
        if response.status_code == 204:
            return True
        else:
            response.raise_for_status()

    def delete_users(self, project_id:str, users:list[dict]):
        """
        Deletes a list of users from a project.

        Args:
            project_id (str): A project UUID
            users (list[dict]): A list of user dictionaries with user e-mails to delete from the project.
        """
        user_emails_set = set([user.get('email') for user in users])
        project_users = self.get_users(project_id)
        users_to_delete = [pu for pu in project_users if pu.get('email') in user_emails_set]

        for user in users_to_delete:
            print(f"Deleting user {user.get('email')} from project {project_id}")
            try:
                self.delete(project_id, user.get('id'))
            except:
                time.sleep(60)
                self.delete(project_id, user.get('id'))

